/*

 normal.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "normal.h"
#include "driver.h"
#include "path.h"
#include "crc8.h"
#include "partition.h"
#include "log.h"
#include "arguments.h"
#include "buffer.h"
#include "ideservd.h"
#ifdef __MINGW32__
#define lstat stat
#endif
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

static void writehex(char *msg, unsigned char c) {
    const char *hex = "0123456789ABCDEF";
    msg[0] = hex[c >> 4];
    msg[1] = hex[c & 15];
    msg[2] = ' ';
}

static inline int send_trailer(int usecrc, const Driver *driver, const Arguments *arguments) {
    if (arguments->mode != M_ETHERNET) {
        if (usecrc) {
            driver->sendb(crc_get());
        }
        driver->sendb(0x5a);
    }
    return driver->flush();
}

static int check_trailer(const char *mode, int usecrc, const Driver *driver) {
    int fr, crcerr = 0;
    if (usecrc) {
        driver->getb();
        crcerr = crc_get();
    }
    fr = driver->getb();
    if (driver->done()) return 1;
    if (fr != 0x5a) {
        seterror(41, 1);
        log_printf("%s: Frame error", mode);
        return 1;
    }
    if (crcerr) {
        seterror(42, 1);
        log_printf("%s: CRC error", mode);
        return 1;
    }
    return 0;
}

int statuserror(const Driver *driver, const char *errorbuff, const Arguments *arguments, int usecrc) {
    Petscii cmd[256];
    unsigned char fp;
    int r;

    driver->getb();
    fp = 0;
    while (fp < 255) {
        r = driver->getb();
        if (r == EOF) break;
        cmd[fp] = r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp] = 0;

    if (arguments->mode != M_ETHERNET) {
        if (r != 0) {
            if (driver->done()) return 1;
            seterror(41, 1);
            log_print("Status: Frame error");
            return 1;
        }
        if (check_trailer("Status", usecrc, driver)) return 1;
    } else {
        if (driver->done()) return 1;
        if (arguments->mode != M_ETHERNET && r != 0) {
            seterror(41, 1);
            log_print("Status: Frame error");
            return 1;
        }
    }

    if (arguments->debug) {
        char msg[16 * 3];
        int i = 0;
        log_print("Command:");
        while (cmd[i]) {
            writehex(msg + (i & 15) * 3, cmd[i]);
            i++;
            if (!(i & 15)) {
                msg[16 * 3 - 1] = '\0';
                log_print(msg);
            }
        }
        if (i & 15) {
            msg[(i & 15) * 3 - 1] = '\0';
            log_print(msg);
        }
    }
    commandchannel(cmd);

    if (usecrc) crc_clear(0);
    fp = strlen(errorbuff);
    driver->sendb(fp);
    if (arguments->mode == M_ETHERNET) driver->sendb(fp);
    driver->sendbytes(fp, (unsigned char *)errorbuff);
    return send_trailer(usecrc, driver, arguments) != 0;
}

int openfile(const Driver *driver, Buffer *buff, const Arguments *arguments, int usecrc) {
    Petscii cmd[256];
    unsigned char status = 0x80;
    unsigned char mode = 'R';
    unsigned char channel, fp;
    unsigned int length = 0;
    Buffer *buffer;
    int r;

    channel = driver->getb() & 0x0f;
    buffer = &buff[channel];

    fp = 0;
    while (fp < 255) {
        r = driver->getb();
        if (r == EOF) break;
        cmd[fp] = r;
        if (!r) break;
        fp++;
    }
    cmd[fp] = 0;

    if (arguments->mode != M_ETHERNET) {
        if (r != 0) {
            if (driver->done()) return 1;
            seterror(41, 1);
            log_print("Open: Frame error");
            return 1;
        }
        if (check_trailer("Open", usecrc, driver)) return 1;
    } else {
        if (driver->done()) return 1;
        if (arguments->mode != M_ETHERNET && r != 0) {
            seterror(41, 1);
            log_print("Open: Frame error");
            return 1;
        }
    }

    if (arguments->debug) {
        char msg[16 * 3];
        int i = 0;
        log_print("Open:");
        while (cmd[i]) {
            writehex(msg + (i & 15) * 3, cmd[i]);
            i++;
            if (!(i & 15)) {
                msg[16 * 3 - 1] = '\0';
                log_print(msg);
            }
        }
        if (i & 15) {
            msg[(i & 15) * 3 - 1] = '\0';
            log_print(msg);
        }
    }
    if (buffer->file) {
        fclose(buffer->file);
        buffer->file = NULL;
    }
    buffer->mode = CM_CLOSED;
    buffer->filepos = 0;

    if (channel == 15) {
        buffer->mode = CM_ERR;
    } else if (cmd[0] == '$') {
        int partition;
        char outpath[1000];
        const Petscii *outname;
        partition_t outpart = 0;

        buffer->mode = CM_DIR;
        buffer->length = buffer->pointer = 0;
        partition = !strncmp((char *)&cmd[1], "=P", 2);

        outname = resolv_path(cmd + 1, outpath, &outpart, arguments->nameconversion);
        if (arguments->verbose) log_printf("Open [%d]: \"$%s\"", channel, outpath);

        if (channel == 1) {
            status = 0x80 | 70; goto vege; //no channel
        } else if (channel > 1) {
            Directory *directory = directory_open(outpath, arguments->nameconversion, 3);
            if (directory == NULL) {
                status = 0x80 | errtochannel15(1);
                log_printf("Open: Couldn't open the directory \"%s\": %s(%d)", outpath, strerror(errno), errno);
                goto vege;
            }
            buffer_rawdir(buffer, directory);
            status = 0x80 | errtochannel15(directory_close(directory));
        } else if (partition) {
            buffer_partition(buffer);
            status = 0x80 | errtochannel15(0);
        } else {
            Directory *directory = directory_open(outpath, arguments->nameconversion, 2);

            if (directory == NULL) {
                status = 0x80 | errtochannel15(1);
                log_printf("Open: Couldn't open the directory \"%s\": %s(%d)", outpath, strerror(errno), errno);
                goto vege;
            }
            buffer_cookeddir(buffer, directory, outpart, outname);
            status = 0x80 | errtochannel15(directory_close(directory));
        }
        memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);
        length = buffer->length;
    } else {
        Petscii name[17], type[4]={'*',0};
        const Petscii *outname;
        unsigned char overwrite;
        char lname[1000], outpath[1020];
        int found = 0;
        int b;
        Directory_entry dirent;
        Directory *directory;

        overwrite = 0;
        for (b = 0; cmd[b]; b++) {
            if (cmd[b] != ':') continue;
            if (cmd[0] == '@') overwrite = 1;
            break;
        }

        outname = resolv_path(cmd, outpath, NULL, arguments->nameconversion);
        if (outname == NULL) {
            status = 0x80 | 39; goto vege;
        }
        if (channel == 1) strcpy((char *)type, "PRG");
        convertfilename(outname, ',', name, type, &mode);

        directory = directory_open(outpath, arguments->nameconversion, 0);
        if (directory == NULL) {
            status = 0x80 | errtochannel15(1);
            log_printf("Open: Couldn't open the directory \"%s\": %s(%d)", outpath, strerror(errno), errno);
            goto vege;
        }

        while (directory_read(directory, &dirent))
        {
            struct stat buf;

            if ((dirent.attrib & (A_ANY | A_CLOSED)) != (A_CLOSED | A_NORMAL)) continue;

            if (!matchname(dirent.name, name)) continue;
            if (!matchname(dirent.filetype, type)) continue;

            if (lstat(directory_path(directory), &buf)) continue;
            length = buf.st_size;

            strncpy(lname, directory_filename(directory), 999);

            found = 1;
            break;
        }
        directory_close(directory);

        if (!found) convertc64name(lname, name, type, arguments->nameconversion);

        if (channel == 0) mode = 'R';
        if (channel == 1) mode = 'W';

        if (outpath[0]) strcat(outpath, "/");
        strcat(outpath, lname);

        if (arguments->verbose) log_printf("Open [%d]: \"%s\" %c", channel, outpath, mode);

        if (!name[0]) {
            status = 0x80 | 34; goto vege;//no filename
        }

        if (mode == 'W') {
            int f = 0;
            length = 0;

            while (lname[f]) {
                if (lname[f] == '*' || lname[f] == '?' || lname[f] == ':' || lname[f] == '=') {
                    status = 0x80 | 33; goto vege;//illegal characters
                }
                f++;
            }

            if (found && !overwrite) {
                status = 0x80 | 63;//exists
            } else {
#ifndef WIN32
                buffer->fd = open(outpath, O_CREAT | O_RDWR | (overwrite ? O_TRUNC : O_EXCL), S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#else
                buffer->fd = open(outpath, O_CREAT | O_RDWR | (overwrite ? O_TRUNC : O_EXCL) | O_BINARY, S_IRUSR | S_IWUSR);
#endif
                if (buffer->fd < 0) status = 0x80 | errtochannel15(1); else {
                    status = 0x80;
                    buffer->file = fdopen(buffer->fd, "wb+");
                    buffer->mode = CM_FILE;
                    buffer->filesize = 0;
                }
            }
        } else {
            if (!found) {
                status = 0x80 | 62;//not found
            } else {
                if (mode == 'R') {
                    buffer->fd = open(outpath, O_RDONLY | O_BINARY, 0);
                    if (buffer->fd < 0) status = 0x80 | errtochannel15(1); else {
                        status = 0x80;//ok
                        buffer->file = fdopen(buffer->fd, "rb");
                        buffer->mode = CM_FILE;
                        buffer->filesize = length;
                    }
                }
                if (mode == 'A' || mode == 'M') {
                    buffer->fd = open(outpath, O_RDWR | O_BINARY, 0);
                    if (buffer->fd < 0) status = 0x80 | errtochannel15(1); else {
                        status = 0x80;//ok
                        buffer->file = fdopen(buffer->fd, "rb+");
                        buffer->mode = CM_FILE;
                        buffer->filesize = length;
                    }
                }
            }
        }
    }
vege:
    {
        unsigned char buf[6];
        if (arguments->mode == M_ETHERNET) {
            buf[0] = length;
            buf[1] = length >> 8;
            buf[2] = length >> 16;
            buf[3] = length >> 24;
            buf[4] = mode;
            buf[5] = status;
        } else {
            if (usecrc) crc_clear(0);
            buf[0] = status;
            buf[1] = mode;
            buf[2] = length;
            buf[3] = length >> 8;
            buf[4] = length >> 16;
            buf[5] = length >> 24;
        }
        driver->sendbytes(sizeof buf, buf);
        return send_trailer(usecrc, driver, arguments) != 0;
    }
}

int closefile(const Driver *driver, Buffer *buff, const Arguments *arguments, int usecrc) {
    unsigned char channel;
    unsigned int length;
    Buffer *buffer;
    unsigned char buf[5];

    driver->getbytes(sizeof buf, buf);
    channel = buf[0] & 0x0f;
    length = (((((buf[4] << 8) | buf[3]) << 8) | buf[2]) << 8) | buf[1];
    if (arguments->mode != M_ETHERNET) {
        if (check_trailer("Close", usecrc, driver)) return 1;
    } else {
        if (driver->done()) return 1;
    }
    buffer = &buff[channel];

    if (arguments->verbose) log_printf("Close [%d] %d byte(s)", channel, length);
    if (channel == 15) buffer->mode = CM_ERR;
    else {
        if (buffer->file) {
            if (buffer->mode == CM_FILE && length >= buffer->filesize) {
                fflush(buffer->file);
                if (ftruncate(buffer->fd, length)) {
                    log_printf("Close: Couldn't truncate: %s(%d)", strerror(errno), errno);
                }
            }
            fclose(buffer->file);
        }
        buffer->mode = CM_CLOSED; buffer->file = NULL;
        if (buffer->data) {
            free(buffer->data);
            buffer->data = NULL;
        }
    }
    if (usecrc) crc_clear(0);
    driver->sendb(0x80);
    if (arguments->mode == M_ETHERNET) driver->sendb(0x80);
    return send_trailer(usecrc, driver, arguments) != 0;
}

int readfile(const Driver *driver, Buffer *buff, const Arguments *arguments, int usecrc) {
    unsigned char channel, sectors;
    struct timeval start, end;
    unsigned long long l = 0, l2;
    unsigned int address;
    Buffer *buffer;
    int bytes = 0;
    unsigned char buf[5];

    driver->getbytes(sizeof buf, buf);
    channel = buf[0] & 0x0f;
    address = (((buf[3] << 8) | buf[2]) << 8) | buf[1];
    sectors = buf[4];
    if (arguments->mode != M_ETHERNET) {
        if (check_trailer("Read", usecrc, driver)) return 1;
    } else {
        if (driver->done()) return 1;
    }
    buffer = &buff[channel];

    if (arguments->verbose) log_printf("Read [%d] %06x %d sector(s)", channel, address, sectors);
    if (sectors > 128 || sectors == 0 || (arguments->mode == M_ETHERNET && sectors > 2)) {
        log_print("Read: Invalid sector count");
        goto error;
    }
    if (buffer->mode != CM_DIR && buffer->mode != CM_ERR && (buffer->mode != CM_FILE || !buffer->file)) {
        log_print((buffer->mode == CM_CLOSED) ? "Read: Channel not open" : "Read: Not readable");
        if (usecrc) crc_clear(0);
        driver->sendb(0x80 | 70);
        if (arguments->mode == M_ETHERNET) driver->sendb(0x80 | 70);
        return send_trailer(usecrc, driver, arguments) != 0;
    }

    if (buffer->mode == CM_FILE) {
        if (buffer->data == NULL) {
            buffer->data = (unsigned char *)malloc(65536);
            if (buffer->data == NULL) {
                log_print("Read: Out of memory");
                goto error;
            }
        }
        if (buffer->filepos != address && fseek(buffer->file, address << 8, SEEK_SET)) {
            buffer->filepos = -1;
            log_printf("Read: Couldn't seek: %s(%d)", strerror(errno), errno);
        error:
            if (usecrc) crc_clear(0);
            driver->sendb(0x80 | 23);
            if (arguments->mode == M_ETHERNET) driver->sendb(0x80 | 23);
            return send_trailer(usecrc, driver, arguments) != 0;
        }
        buffer->filepos = address;
        clearerr(buffer->file);
    }

    if (arguments->mode == M_ETHERNET) {
        const unsigned char *data;
        if (buffer->mode == CM_FILE) {
            size_t itt = fread(buffer->data, 1, sectors * 512, buffer->file);
            buffer->filepos += ((itt + 511) >> 9) * 2;
            if (ferror(buffer->file)) {
                buffer->filepos = -1;
                log_printf("Read: Couldn't read: %s(%d)", strerror(errno), errno);
                driver->sendb(0x80 + 23);
                driver->sendb(0x80 + 23);
                if (send_trailer(usecrc, driver, arguments)) return 1;
            }
            if (itt < sectors * 512) memset(buffer->data + itt, 0, sectors * 512 - itt);
            data = buffer->data;
        } else {
            data = buffer->data + buffer->pointer;
        }
        driver->sendbytes(sectors * 512, data);
        driver->sendb(0x80);
        driver->sendb(0x80);
        if (send_trailer(usecrc, driver, arguments)) return 1;
        buffer->pointer += sectors * 512;
    } else {
        if (usecrc) crc_clear(0);
        driver->sendb(0x80);
        if (send_trailer(usecrc, driver, arguments)) return 1;
        if (arguments->verbose) {
            bytes = sectors * 512;
            gettimeofday(&start, NULL);
        }
        while (sectors) {
            int err = 0;
            const unsigned char *data;
            if (buffer->mode == CM_FILE) {
                size_t itt = fread(buffer->data, 1, 512, buffer->file);
                buffer->filepos += ((itt + 511) >> 9) * 2;
                err = ferror(buffer->file);
                if (err) {
                    buffer->filepos = -1;
                    log_printf("Read: Couldn't read: %s(%d)", strerror(errno), errno);
                }
                if (itt < 512) memset(buffer->data + itt, 0, 512 - itt);
                data = buffer->data;
            } else {
                data = buffer->data + buffer->pointer;
            }
            if (arguments->mode == M_RS232 || arguments->mode == M_RS232S) {
                int fr = driver->getb();
                if (driver->done()) return 1;
                if (fr != 0x5a) {
                    seterror(41, 1);
                    log_print("Read: Frame error");
                    return 1;
                }
            }
            if (usecrc) crc_clear(0);
            driver->sendbytes(512, data);
            driver->sendb(err ? 0x80 + 23 : 0x80);
            if (send_trailer(usecrc, driver, arguments)) return 1;
            if (err) break;
            buffer->pointer += 512;
            sectors--; l += 512;
        }
        if (arguments->verbose) {
            gettimeofday(&end, NULL);
            l2 = ((end.tv_sec - start.tv_sec) * 1000000 + end.tv_usec - start.tv_usec);
            if (l2) {
                l = l * 1000000 / l2;
                log_printf("Read: Sent %d bytes at %u bytes/seconds", bytes, (unsigned int)l);
            } else {
                log_printf("Read: Sent %d bytes", bytes);
            }
        }
    }
    return 0;
}

int writefile(const Driver *driver, Buffer *buff, const Arguments *arguments, int usecrc) {
    unsigned char channel, sectors;
    struct timeval start, end;
    unsigned long long l = 0, l2;
    unsigned int address;
    Buffer *buffer;
    int fr, bytes = 0;
    unsigned char buf[5];

    driver->getbytes(sizeof buf, buf);
    channel = buf[0] & 0x0f;
    address = (((buf[3] << 8) | buf[2]) << 8) | buf[1];
    sectors = buf[4];
    if (arguments->mode != M_ETHERNET) {
        if (check_trailer("Write", usecrc, driver)) return 1;
    } else {
        if (driver->done()) return 1;
    }
    buffer = &buff[channel];

    if (arguments->verbose) log_printf("Write [%d] %06x %d sector(s)", channel, address, sectors);
    if (sectors > 128 || sectors == 0 || (arguments->mode == M_ETHERNET && sectors > 2)) {
        log_print("Write: Invalid sector count");
        goto error;
    }
    if (buffer->mode != CM_FILE || buffer->file == NULL) {
        log_print("Write: Not writeable");
    error:
        if (usecrc) crc_clear(0);
        driver->sendb(0x80 + 25);
        if (arguments->mode == M_ETHERNET) driver->sendb(0x80 + 25);
        return send_trailer(usecrc, driver, arguments) != 0;
    }
    if (buffer->data == NULL) {
        buffer->data = (unsigned char *)malloc(65536);
        if (buffer->data == NULL) {
            log_print("Write: Out of memory");
            goto error;
        }
    }
    if (buffer->filepos != address && fseek(buffer->file, address << 8, SEEK_SET)) {
        buffer->filepos = -1;
        log_printf("Write: Couldn't seek: %s(%d)", strerror(errno), errno);
        goto error;
    }
    buffer->filepos = address;
    if (arguments->mode == M_ETHERNET) {
        driver->getbytes(512 * sectors, buffer->data);
        if (driver->done()) return 1;
        if (fwrite(buffer->data, 1, 512 * sectors, buffer->file) == 512 * sectors && fflush(buffer->file) == 0) {
            buffer->filepos += sectors * 2;
            driver->sendb(0x80);
            driver->sendb(0x80);
        } else {
            buffer->filepos = -1;
            log_printf("Write: Couldn't write: %s(%d)", strerror(errno), errno);
            driver->sendb(0x80 + 25);
            driver->sendb(0x80 + 25);
        }
        if (send_trailer(usecrc, driver, arguments)) return 1;
    } else {
        int err = 0;
        if (usecrc) crc_clear(0);
        driver->sendb(0x80);
        if (send_trailer(usecrc, driver, arguments)) return 1;

        if (arguments->verbose) {
            bytes = sectors * 512;
            gettimeofday(&start, NULL);
        }
        while (sectors) {
            int crcerr = 0;
            if (usecrc) crc_clear(0);
            driver->getbytes(513, buffer->data);
            if (usecrc) {
                crcerr = crc_get();
                fr = driver->getb();
            } else {
                fr = buffer->data[512];
            }
            if (driver->done()) return 1;
            if (fr != 0x5a) {
                seterror(41, 1);
                log_print("Write: Frame error");
                return 1;
            }
            if (crcerr) {
                seterror(42, 1);
                log_print("Write: CRC error");
                return 1;
            }
            if (err == 0 && fwrite(buffer->data, 1, 512, buffer->file) != 512) err = errno;
            buffer->filepos += 2;
            sectors--; l += 512;
            if (sectors == 0 || arguments->mode == M_RS232 || arguments->mode == M_RS232S) {
                if (err == 0 && fflush(buffer->file) != 0) err = errno;
                if (err != 0) {
                    buffer->filepos = -1;
                    log_printf("Write: Couldn't write: %s(%d)", strerror(err), err);
                }
                if (usecrc) crc_clear(0);
                driver->sendb(err ? 0x80 + 25 : 0x80);
                if (send_trailer(usecrc, driver, arguments)) return 1;
                if (err) break;
            }
        }
        if (arguments->verbose) {
            gettimeofday(&end, NULL);
            l2 = ((end.tv_sec - start.tv_sec) * 1000000 + end.tv_usec - start.tv_usec);
            if (l2) {
                l = l * 1000000 / l2;
                log_printf("Write: Received %d bytes at %u bytes/seconds", bytes, (unsigned int)l);
            } else {
                log_printf("Write: Received %d bytes", bytes);
            }
        }
    }
    return 0;
}

